#!/bin/sh
#  Copyright 2008 Nicholas A. Schembri State College PA USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see
#    <http://www.gnu.org/licenses/>.

# Thank you Voyage Linux for the idea, http://voyage.hk/ Great job on release 0.5
#
# Tested with 8.04.1
# tested with 9.10
# tested with debian live 6.0.1
#
# ****************************************************************************************
# 
#                                 Change log
#
# 2008.08.01  Added debugging comments in "drop to a shell" section. grub option aufs=tmpfs-debug will stop the init script.
#             reviewed *********** fix fstab on tmpfs ******************
#             rootaufs failed when system was booted with /dev/xxx and fstab had uuid= info.
#             BlaYO pointed out the best and simplest solution was to use grep -v. Grep replaces a sed one liner.
#             Add the comment block to fstab
#             
# 2009.12.07 Corrected issue caused by Apparmor.
#            Name changed to __rootaufs.
# 
# 2011.08.19 Changed if condition to avoid issue (sh: bad number) when $aufsdebug is not set.
#            Now checks exists apparmor before delete.
#
# 2011.08.20 For work correctly with Debian Live 6.0.1 (http://live.debian.net/) two lines must be removed from rc.local modification part:
#             'mount -f  /ro'
#             'echo aufs-tmpfs /rw tmpfs rw 0 0 >>/etc/mtab'
#





case $1 in
prereqs)
    echo ""
    exit 0
    ;;
esac

export overlay
export overlay_driver
AREA=
NOJRE=

for x in $(cat /proc/cmdline); do
    case $x in
    root=*)
        ROOTNAME=${x#root=}
	case $ROOTNAME in
	AREA=*)
		AREA="${ROOTNAME#AREA=}"
		;;
	esac
        ;;
    overlay=*)
        overlay=${x#overlay=}
        case $overlay in
        tmpfs-debug)
            overlay=tmpfs
            overlaydebug=1
            ;;
        esac
        ;;
    overlay_driver=*)
	overlay_driver=${x#overlay_driver=}
	;;
    nojre)
        NOJRE=1
        ;;
    esac
done

mount_overlayfs() {
    local lower=$1
    local upper=$2
    local name=$3
    local mntpoint=$4

    mount -t overlayfs -o lowerdir=${lower},upperdir=${upper} $name $mntpoint > /dev/null 2>&1
    echo $?
}

mount_aufs() {
    local lower=$1
    local upper=$2
    local name=$3
    local mntpoint=$4

    mount -t aufs -o dirs=${upper}:${lower}=rr $name $mntpoint > /dev/null 2>&1
    echo $?
}

if [ "$overlay_driver" = "" ]; then
    overlay_driver=overlayfs
fi

if [ "$overlay_driver" = "overlayfs" ]; then
    MOUNT=mount_overlayfs
elif [ "$overlay_driver" = "aufs" ]; then
    MOUNT=mount_aufs
else
    echo "'overlay_driver' error!"
    exit 0
fi

if [ "$overlay" != "tmpfs" ] && [ "$overlay" != "mmc" ]; then
    #not set in boot loader
    #I'm not loved. good bye
    exit 0
fi


# This is a simple overview of the steps needed to use aufs on the root file system and see the /rw and /ro  branches.
# initramfs init-botton script 
# move the root file system to aufs/unionfs readonly /ro
# root is mounted on ${rootmnt}
# create tmpfs on /rw
# create a aufs using /ro and /rw
# put some files on the tmpfs to fix mtab and fstab 
# move aufs to rootmnt to finish the init process.
# No changes to the root file system are made by this script.
#
#  Why!
#  This will allow you to use a usb flash drive and control what is written to the drive.
#  no need to rebuild the squashfs file just to add a program. 
#  boot to single user mode.  The system works the way you expect. boot aufs=tmpfs and no changes are written to the flash.
#  run ubuntu on an eeePC .  

# Install 
# Install ubuntu 8.04 Hardy. Hardy has aufs installed by default
# apt-get update
# apt-get dist-upgrade
# apt-get install aufs-tools
# echo aufs >> /etc/initramfs-tools/modules
# put this file in /etc/initramfs-tools/scripts/init-bottom/rootaufs
# chmod 0755 rootaufs
# # clean up menu.lst 
# update-grub
# update-initramfs -u
# vi /boot/grub/menu.lst
# add aufs=tmpfs to the default entry. 
# do not add this line to single user mode.
# boot to single user mode in order to install software. 
# note: if your home account is on the root file system, your files are in ram and not saved.
# 


echo 
echo "       root-overlay:  Setting up ${overlay_driver} on ${rootmnt} as root file system "
echo 

if ! cat /proc/filesystems | grep $overlay_driver > /dev/null 2>&1 ; then
	modprobe -q --use-blacklist $overlay_driver

	if [ $? -ne 0 ]; then
	    echo    root-overlay error:      Failed to load ${overlay_driver}.ko
	    exit 0
	fi
fi

#make the mount points on the init root file system
mkdir /overlay
mkdir /rw
mkdir /rw.tmpfs
mkdir /ro

use_persist=0

# mount the temp file system and move real root out of the way
mount -t tmpfs overlay-tmpfs /rw.tmpfs
if [ "$overlay" = "tmpfs" ]; then
    mount -t tmpfs overlay-persist /rw
else
    MMCDEV=/dev/mmcblk0p1
    MMCFS=`blkid /dev/mmcblk0p1 |awk '{print $NF}'|sed 's/TYPE="\([0-9a-z]\+\)"/\1/'`

    echo "MMC Filesystem: $MMCFS detected."

    mkdir -p /rw.tmpfs/mnt/mmc

    if [ "$MMCFS" = "vfat" ]; then
	fsck.vfat -a $MMCDEV > /dev/null 2>&1
    elif [ "$MMCFS" = "ext4" ]; then
	fsck.ext4 -y $MMCDEV > /dev/null 2>&1
    else
	echo "filesystem $MMCFS not supported."
	exit 0
    fi

    mount -t $MMCFS $MMCDEV /rw.tmpfs/mnt/mmc > /dev/null 2>&1

    if [ "$?" = "0" ]; then
	loopback_file=/rw.tmpfs/mnt/mmc/overlay.img

	mount_ok=0
	if [ -f $loopback_file ]; then
	    mount -o loop -t ext4 $loopback_file /rw > /dev/null 2>&1
	    if [ "$?" = "0" ]; then
		mount_ok=1
	    else
		echo "loopback mount failed, initialize required."
		losetup -d /dev/loop0 > /dev/null 2>&1
	    fi
	else
	    for retry in 1 2 3 4 5 6; do
		free_KB=`df | grep -F /mnt/mmc | awk '{print $4}'`
		if [ "$free_KB" != "" ]; then
		    break
		fi
		sleep 0.5
	    done

	    develfile=`find /rw.tmpfs/mnt/mmc -maxdepth 1 -name "[Dd][Ee][Vv][Ee][Ll]" | head -1`
	    if [ "$develfile" != "" ]; then
		echo "'devel' file found."
		content=`cat $develfile`
		size_req=`echo $(($content))` 2> /dev/null

		if [ "$?" != "0" ]; then
		    echo "equation error"
		elif [ "$size_req" -gt "$free_KB" ]; then
		    echo "requested size over ($size_req [KiB] > $free_bytes [KiB])"
		    size_req=$free_KB
		fi
	    else
		echo "'devel' file not found, ???"
	    fi

	    if [ "$size_req" = "" ]; then
		size_req=0
	    else
		# KiB -> MiB
		size_req=$((size_req/1024))

		size4G=$((4*1024))
		if [ "$MMCFS" = "vfat" ] && [ "$size_req" -gt "$size4G" ]; then
		    echo "*** VFAT support max 4GiB"
		    size_req=$size4G
		fi
	    fi

	    free_MB=$((free_KB/1024))
	    imgsize=$((free_MB*8/10))

	    if [ "$size_req" != "0" ]; then
		if [ "$size_req" -lt "$imgsize" ]; then
		    imgsize=$size_req
		elif [ "$size_req" -gt "$free_MB" ]; then
		    imgsize=$free_MB
		fi
	    else
		size2G=$((2*1024))

		if [ "$imgsize" -gt "$size2G" ]; then
		    imgsize=$size2G
		fi
	    fi
	fi

	if [ "$mount_ok" != "1" ]; then
	    if [ ! -f $loopback_file ]; then
		# need long time, stop watchdog timer
		echo V > /dev/watchdog

		echo -n "create loopback image file ($imgsize [MiB]) ..."
		# create loopback image
		dd if=/dev/zero of=$loopback_file bs=1048576 count=$imgsize
		echo "done."
	    fi

	    echo -n "creating ext4 filesystem on image file..."
	    mkfs.ext4 -m 0 -F $loopback_file > /dev/null 2>&1
	    echo "done."

	    mount -o loop -t ext4 $loopback_file /rw > /dev/null 2>&1
	    if [ "$?" = "0" ]; then
		mount_ok=1
	    fi
	fi

	if [ "$mount_ok" = "1" ]; then
	    use_persist=1
	    # PowerLED: Orange
	    echo 128 > /sys/class/leds/power-stat\:r/brightness
	else
	    # cleanup
	    umount /rw.tmpfs/mnt/mmc
	fi
    fi

    if [ "$use_persist" != "1" ]; then
	echo "mount mmc failed, fallback to tmpfs."
	mount -t tmpfs aufs-persist /rw
    fi
fi

# mount modules, JRE
if [ "$AREA" = "0" ] || [ "$AREA" = "1" ]; then
    # mount kernel module area to /lib/modules
    MTD_KMOD=`cat /proc/mtd | grep kmod.$AREA | cut -d ':' -f 1`

    if [ -n "$MTD_KMOD" ]; then
	MTD_KMOD="/dev/mtdblock${MTD_KMOD#mtd}"
	IMGFILE=/rw.tmpfs/kmod.img

	#cp $MTD_KMOD $IMGFILE
	tar xfO $MTD_KMOD | cat > $IMGFILE

	mount -t squashfs $IMGFILE ${rootmnt}/lib/modules > /dev/null 2>&1
	if [ "$?" != "0" ]; then
	    rm $IMGFILE
	fi
    fi

    # mount JRE to /usr/lib/jvm
    if [ "$NOJRE" = "" ]; then
	JVM_DIR=/usr/lib/jvm
	MTD_JRE=`cat /proc/mtd | grep -F opt | cut -d ':' -f 1`

	if [ -n "$MTD_JRE" ]; then
	    MTD_JRE="/dev/mtdblock${MTD_JRE#mtd}"
	    IMGFILE=/rw.tmpfs/jre.img

	    #cp $MTD_JRE $IMGFILE
	    tar xfO $MTD_JRE | cat > $IMGFILE

	    mount -t squashfs $IMGFILE ${rootmnt}${JVM_DIR} > /dev/null 2>&1
	    if [ "$?" != "0" ]; then
		rm $IMGFILE
	    fi
	fi
    fi
fi


mount --move ${rootmnt} /ro
if [ $? -ne 0 ]; then
    echo    root-overlay error:    ${rootmnt}  failed to move to /ro
    exit 0
fi


res=`$MOUNT /ro /rw overlay /overlay`
if [ $res -ne 0 ]; then
    echo    root-overlay error:      Failed to mount /overlay files system
    exit 0
fi

mkdir -p /rw.tmpfs/var/log
$MOUNT /ro/var/log /rw.tmpfs/var/log overlay.log /overlay/var/log

mkdir -p /rw/lib/modules
$MOUNT /ro/lib/modules /rw/lib/modules overlay.kmod /overlay/lib/modules

if [ "$NOJRE" = "" ]; then
    mkdir -p /rw/usr/lib/jvm
    $MOUNT /ro/usr/lib/jvm /rw/usr/lib/jvm overlay.jvm /overlay/usr/lib/jvm
fi


#test for mount points on aufs file system
OVERLAY_BASE=/overlay
[ -d ${OVERLAY_BASE}/ro ] || mkdir ${OVERLAY_BASE}/ro
[ -d ${OVERLAY_BASE}/rw ] || mkdir ${OVERLAY_BASE}/rw
[ -d ${OVERLAY_BASE}/rw.tmpfs ] || mkdir ${OVERLAY_BASE}/rw.tmpfs

# the real root file system is hidden on /ro of the init file system.  move it to /ro 
mount --move /ro ${OVERLAY_BASE}/ro
if [ $? -ne 0 ]; then
    echo    root-overlay error:      Failed to move /ro ${OVERLAY_BASE}/ro
    exit 0
fi

# overlay file system is hidden on /rw
mount --move /rw ${OVERLAY_BASE}/rw
if [ $? -ne 0 ]; then
    echo    root-overlay error:      Failed to move /rw ${OVERLAY_BASE}/rw
    exit 0
fi

# tmpfs file system is hidden on /rw.tmpfs
mount --move /rw.tmpfs ${OVERLAY_BASE}/rw.tmpfs
if [ $? -ne 0 ]; then
    echo    root-overlay error:      Failed to move /rw ${OVERLAY_BASE}/rw.tmpfs
    exit 0
fi

# mount UBIFS
ubi_overlay_dir=${OVERLAY_BASE}/rw.tmpfs/.overlay
mkdir $ubi_overlay_dir
mount -t ubifs ubi0:overlay $ubi_overlay_dir

# extract saved contents
restore_overlays() {
    filename=`basename $1`
    echo -n "extracting $filename to $2 ... "

    tar axf $1 -C $2 > /dev/null 2>&1

    if [ "$?" = "0" ]; then
        echo "succeeded."
    else
        echo "failed."
    fi
}

# check if 'cleanboot=1' string exists?
cleanboot=0
for x in `cat /proc/cmdline`; do
    if [ "$x" = "cleanboot=1" ]; then
        cleanboot=1
        break
    fi
done

if [ "$cleanboot" = "0" ]; then
    if [ "$use_persist" = "0" ]; then
	for f in `find $ubi_overlay_dir -name overlays*.tar.*`; do
            restore_overlays $f /overlay
	done
    else
	# restore /var/log only
	for f in `find $ubi_overlay_dir -name overlays_log.tar.*`; do
           restore_overlays $f /overlay
	done
    fi
else
    echo "don't extract overlay archives..."
fi



#*********** fix fstab on tmpfs ******************
# test for /dev/sdx 
# this is not on the real file system.  This is created on the tmpfs each time the system boots.
# The init process will try to mount the root filesystem listed in fstab. / and swap must be removed.  
# the root file system must be mounted on /ro not on /

if [ "0$overlaydebug" -eq 1 ]; then
    echo  "   root-overlay debug:    Remove the root file system and swap from fstab "
    echo 
    echo 
    echo  "         ROOTNAME $ROOTNAME "
    echo  "         resume   $resume   "
    echo 
    echo  '     BlaYO pointed out that grep can be used to quickly remove '
    echo  '      the root file system from fstab. '
    echo 
    echo  '     Thank you BlaYO for the debug info.'
    echo

fi
# old code
# I'm sure that sed can do this in one step but I want to correct on the rootname  not matching the root in fstab.
#cat /overlay/ro/etc/fstab|sed -e s/$ROOTNAME/\#$ROOTNAME/ -e s/$resume/\#$resume/ >/overlay/etc/fstab  

#Add the comment block to fstab
cat <<EOF >/overlay/etc/fstab
#
#   RootAufs has mounted the root file system in ram
#
#  This fstab is in ram and the real fstab can be found /ro/etc/fstab
#  the root file system ' / ' has been removed.
#  All Swap files have been removed. 
#

EOF

#remove root and swap from fstab
cat ${OVERLAY_BASE}/ro/etc/fstab|grep -v ' / ' | grep -v swap >>/overlay/etc/fstab
if [ $? -ne 0 ]; then
    echo    root-overlay error:      Failed to create /overlay/etc/fstab
    #exit 0
fi




# add the read only file system to fstab
#ROOTTYPE=$(/lib/udev/vol_id -t ${ROOT})
ROOTTYPE=$(cat /proc/mounts|grep ${ROOT}|cut -d' ' -f3)
ROOTOPTIONS=$(cat /proc/mounts|grep ${ROOT}|cut -d' ' -f4)
echo ${ROOT} /.overlay/ro $ROOTTYPE $ROOTOPTIONS 0 0 >>/overlay/etc/fstab


# S22mount on debian systems is not mounting  /ro correctly after boot
# add to rc.local to correct what you see from df
#replace last case of exit with #exit
#cat /overlay/ro/etc/rc.local|sed 's/\(.*\)exit/\1\#exit/' >/overlay/etc/rc.local  
#echo mount -f  /ro >>/overlay/etc/rc.local 

# add back the root file system. mtab seems to be created by one of the init proceses. 
#echo "echo aufs / aufs rw,xino=/rw/.aufs.xino,br:/rw=rw:/ro=ro 0 0 >>/etc/mtab" >>/overlay/etc/rc.local
#echo "echo aufs-tmpfs /rw tmpfs rw 0 0 >>/etc/mtab" >>/overlay/etc/rc.local 
#echo exit 0 >>/overlay/etc/rc.local 

#  Copyright 2008 Joaquín I. Bogado García
#fix para apparmor, se desactiva y listo ( From the lethe project. )
[ -e /scripts/init-bottom/_apparmor ] && rm /scripts/init-bottom/_apparmor
[ -e /overlay/etc/init.d/apparmor ] && rm /overlay/etc/init.d/apparmor


#build remountrw
#echo \#!/bin/sh >/overlay/bin/remountrw
#echo mount -o remount,rw ${ROOT} >>/overlay/bin/remountrw
#chmod 0700 /overlay/bin/remountrw

#build remountro
#echo \#!/bin/sh >/overlay/bin/remountro
#echo mount -o remount,ro ${ROOT} >>/overlay/bin/remountro
#chmod 0700 /overlay/bin/remountro

# This should drop to a shell. (rewrite)
if [ "0$aufsdebug" -eq 1 ]; then
    echo
    echo "   root-overlay debug:    mount --move /overlay ${rootmnt} "
    echo 
    echo '   root-overlay debug:   init will stop here.   '
    echo  
    exit 0
fi

mount --move /overlay ${rootmnt}

exit 0
